// ************************************************************************************************
//
// Volt- und Amperemeter unter Verwendung des
//
// PIC 16F628A
//
// (x) Erfassen des Spannungswertes durch 10Bit-ADC MAX1242
// (x) Erfassen des Stromwertes ber Operationsverstrker und 8Bit-ADC TLC549
// (x) Mittelwertbildung ber 20 Messungen
//
// (x) Ausgabe der Messwerte ber 2x16-Zeichen LCD
//
//
// http://www.jb-electronics.de/html/elektronik/nixies/n_netzteil3.htm
//
// Copyright www.jb-electronics.de
//
// ************************************************************************************************


// ************************************************************************************************
// PIC-Header einbinden
#include <pic.h>


// ************************************************************************************************
// Methodenprototypen
void sendeWert (unsigned char wert);
int getADC_VOLTAGE ();
int getADC_CURRENT ();

void warte ();
void sendeWert (unsigned char Wert);
void initLCD ();
void updateLCD ();
void sendCommand (unsigned char Command);
void writeChar (unsigned char Zeichen, unsigned char Row, unsigned char Column);
void writeCharBuffered (unsigned char Zeichen, unsigned char Row, unsigned char Column);
void writeTextBuffered (unsigned char Text[], unsigned char Row, unsigned char Column, unsigned char Length);
void setPosition (unsigned char Row, unsigned char Column);


// ************************************************************************************************
// Die Textdaten
static bank1 unsigned char Zeile1[16] = {' ', 'U', ' ', ' ', ' ', ' ', 'V', ' ', ' ', 'R', ' ', ' ', ' ', ' ', ' ', ' '};
static bank1 unsigned char Zeile2[16] = {' ', ' ', ' ', ' ', 'I', ' ', ' ', ' ', '.', ' ', 'm', 'A', ' ', ' ', ' ', ' '};


// ************************************************************************************************
// Globale Variablen
static int TimeCounter, WertADC, mean_strom, mean_spannung;
static int Strom, Spannung, Widerstand;
static unsigned char mean_counter, w;


// ************************************************************************************************
// Hauptmethode
void main (void) {


    // ********************************************************************************************
	// Konfigurationsbits setzen
	// Power-Up-Timer, interner Oszillator (4MHz)
	__CONFIG(0x3F70);

	
	// ********************************************************************************************
	// Portspezifikationen

	#define DATA_VOLTAGE	RB1
	#define CS_VOLTAGE		RB2
	#define CLK_VOLTAGE		RB3

	#define DATA_CURRENT	RA4
	#define CS_CURRENT		RA3
	#define CLK_CURRENT		RB0

	#define DATA_REGISTER	RA0
	#define CLK_REGISTER	RA7
	#define STROBE_REGISTER	RA1

	#define RS				RB4
	#define RW				RB5
	#define ENABLE			RB6

	#define R_ANODE_0		RA2
	#define R_ANODE_1		RA6
	#define	R_ANODE_4		RB7

	
	// Tristate von PORTA und PORTB setzen
	TRISA = 0b01010100;
	TRISB = 0b10000010;

	
	// Analogkomparator bei PORTA abschalten
	CMCON = 0b111;


	// ********************************************************************************************
	// TIMER setzen und Interrupt bei Timer-berlauf auslsen
	T0CS = 0;
	PSA = 1;
	PS0 = 1;
	PS1 = 1;
	PS2 = 1;
	T0IE = 1;

	// Interrupts aktivieren
	GIE = 1;


	// ********************************************************************************************
	// Variablen initialisieren
	mean_strom = 0;
	Strom = 0;
	mean_spannung = 0;
	Spannung = 0;


	// ********************************************************************************************
	// LCD initialisieren und warten
	initLCD();
	for (w = 0; w < 200; w++) {
		NOP();
	}


	// ********************************************************************************************
	// Hauptschleife
	while (1) {

	
		// Spannung anzeigen
		writeCharBuffered (48 + Spannung / 100, 1, 4);
		writeCharBuffered (48 + Spannung / 10 - (Spannung / 100) * 10, 1, 5);
		writeCharBuffered (48 + Spannung % 10, 1, 6);
	
	
		// Strom anzeigen
		writeCharBuffered (48 + Strom / 100, 2, 7);
		writeCharBuffered (48 + Strom / 10 - (Strom / 100) * 10, 2, 8);
		writeCharBuffered (48 + Strom % 10, 2, 10);
	
	
		// eingestellten Anodenvorwiderstand anzeigen
		Widerstand = R_ANODE_0 + (R_ANODE_1 << 1) + (R_ANODE_4 << 2);
		if (Widerstand == 0) {
			writeCharBuffered ('n', 1, 12);
			writeCharBuffered ('o', 1, 13);
			writeCharBuffered ('n', 1, 14);
			writeCharBuffered ('e', 1, 15);
		} else if (Widerstand == 1) {
			writeCharBuffered ('4', 1, 12);
			writeCharBuffered ('k', 1, 13);
			writeCharBuffered ('7', 1, 14);
			writeCharBuffered (' ', 1, 15);
		} else if (Widerstand == 2) {
			writeCharBuffered ('1', 1, 12);
			writeCharBuffered ('8', 1, 13);
			writeCharBuffered ('k', 1, 14);
			writeCharBuffered (' ', 1, 15);
		} else if (Widerstand == 3) {
			writeCharBuffered ('4', 1, 12);
			writeCharBuffered ('7', 1, 13);
			writeCharBuffered ('k', 1, 14);
			writeCharBuffered (' ', 1, 15);
		} else if (Widerstand == 4) {
			writeCharBuffered ('1', 1, 12);
			writeCharBuffered ('0', 1, 13);
			writeCharBuffered ('0', 1, 14);
			writeCharBuffered ('k', 1, 15);
		} else if (Widerstand == 5) {
			writeCharBuffered ('2', 1, 12);
			writeCharBuffered ('2', 1, 13);
			writeCharBuffered ('0', 1, 14);
			writeCharBuffered ('k', 1, 15);
		}

		
		// Daten an LCD senden
		updateLCD ();

	
	}

	
}


// ************************************************************************************************
// Interrupt-Routine zum Abfragen des ADCs
static void interrupt isr (void) {


	// ********************************************************************************************
	// ein bisschen Zeit verstrichen?
	if (T0IF) {
	
		TimeCounter++;
		if (TimeCounter >= 10) {

			// Werte holen
			mean_counter++;
			mean_strom += getADC_CURRENT();
			mean_spannung += getADC_VOLTAGE();

			if (mean_counter >= 20) {
				Strom = mean_strom >> 5;
				mean_strom = 0;
				Spannung = mean_spannung >> 6;
				mean_spannung = 0;
				mean_counter = 0;
			}
	
			TimeCounter = 0;

		}

		// Interrupt resetten
		T0IF = 0;

	}

}


// ************************************************************************************************
// sendet einen Wert an das Schieberegister
void sendeWert (unsigned char wert) {

	// Variablen
	int Index;

	STROBE_REGISTER = 0;
	CLK_REGISTER = 0;
	for (Index = 7; Index >= 0; Index--) {
		if (wert & (1 << Index)) {
			DATA_REGISTER = 1;
		} else {
		 	DATA_REGISTER = 0;
		}
		CLK_REGISTER = 1;
		CLK_REGISTER = 0;
	}
	STROBE_REGISTER = 1;
	STROBE_REGISTER = 0;

}


// ************************************************************************************************
// Diese Methode fragt den Spannungs-ADC ab
int getADC_VOLTAGE () {

	// Variablen
	unsigned char Index;
	int adc_wert = 0;
	
	// ADC auswhlen (/ChipSelect = 0)
	CS_VOLTAGE = 0;

	// Zehn Datenbits der Reihe nach reinholen (Start: MSB...LSB)
	for (Index = 0; Index <= 9; Index++) {

		// positive Flanke
		CLK_VOLTAGE = 1;
		NOP();
		CLK_VOLTAGE = 0;
		NOP();

		// Datenbit reinholen
		if (DATA_VOLTAGE) {
			adc_wert |= 1 << (9 - Index);
		}
		NOP();

	}

	// ADC abwhlen
	CS_VOLTAGE = 1;

	return adc_wert;

}


// ************************************************************************************************
// Diese Methode fragt den Strom-ADC ab
int getADC_CURRENT () {

	// Variablen
	unsigned char Index;
	int adc_wert = 0;
	
	// ADC auswhlen (/ChipSelect = 0)
	CS_CURRENT = 0;

	// Acht Datenbits der Reihe nach reinholen (Start: MSB...LSB)
	for (Index = 0; Index <= 7; Index++) {

		// Datenbit reinholen
		if (DATA_CURRENT) {
			adc_wert |= 1 << (7 - Index);
		}
		NOP();

		// positive Flanke
		CLK_CURRENT = 1;
		NOP();
		CLK_CURRENT = 0;
		NOP();

	}

	// ADC abwhlen
	CS_CURRENT = 1;

	return adc_wert;

}


// ************************************************************************************************
// Diese Methode initialisiert das LCD
void initLCD () {

	// warten
	int i;
	ENABLE = 0;
	for (i = 0; i < 5000; i++) {
		NOP();
	}

	// Cursor unsichtbar
	// setCursor(1, 0, 0);
	sendCommand(8 + (1 << 2));

	// 2zeilig
	// setFunction (1, 1, 1);
	sendCommand(32 + (1 << 4) + (1 << 3) + (1 << 2) + 2 + 1);

	// Cursor auf Ausgangsposition
	// setCursorHome ();
	sendCommand(3);

	// Display lschen
	// clearDisplay();
	sendCommand(1);
	for (w = 0; w < 200; w++) {
		NOP();
	}

}


// ************************************************************************************************
// Diese Methode sendet ein Kommando an das LCD
void sendCommand (unsigned char Command) {

	RW = 0;
	RS = 0;
	sendeWert (Command);
	ENABLE = 1;
	NOP();
	ENABLE = 0;
	warte ();

}


// ************************************************************************************************
// Diese Methode sendet ein Zeichen in den PIC-Zeichenpuffer
void writeCharBuffered (unsigned char Zeichen, unsigned char Row, unsigned char Column) {

	if (Row == 1) {
		Zeile1[Column - 1] = Zeichen;
	} else if (Row == 2) {
		Zeile2[Column - 1] = Zeichen;
	}

}


// ************************************************************************************************
// Diese Methode sendet ein Zeichen direkt an das LCD
void writeChar (unsigned char Zeichen, unsigned char Row, unsigned char Column) {

	// Cursor-Position setzen
	setPosition (Row, Column);

	// Zeichen senden
	RW = 0;
	RS = 1;
	sendeWert (Zeichen);
	ENABLE = 1;
	NOP();
	ENABLE = 0;
	warte ();

}


// ************************************************************************************************
// Diese Methode schreibt eien Text an das Display
void writeTextBuffered (unsigned char Text[], unsigned char Row, unsigned char Column, unsigned char Length) {

	unsigned char t;
	for (t = 0; t < Length; t++) {
		writeCharBuffered (Text[t], Row, Column + t);
	}

}


// ************************************************************************************************
// Diese Methode schreibt alle Daten aus dem PIC-Zeichenpuffer in das LCD
void updateLCD () {

	unsigned char t;
	for (t = 1; t <= 16; t++) {
		writeChar (Zeile1[t - 1], 1, t);
		writeChar (Zeile2[t - 1], 2, t);
	}

}


// ************************************************************************************************
// Diese Methode platziert den Cursor
void setPosition (unsigned char Row, unsigned char Column) {

	if (Row == 1) {
		// setAddressDDRAM(Column - 1);
		sendCommand (127 + Column);
	} else {
		// setAddressDDRAM(64 + Column - 1);
		sendCommand (191 + Column);
	}

}


// ************************************************************************************************
// Diese Methode wartet kurz, um Signale gltig werden zu lassen
void warte () {

	for (w = 0; w < 5; w++) {
		NOP();
	}

}

